package bench;

import java.io.File;

import com.hp.hpl.jena.util.FileManager;

import com.hp.hpl.jena.sparql.util.Utils;

import arq.cmdline.ArgDecl;
import arq.cmdline.CmdArgModule;
import arq.cmdline.CmdGeneral;

import benchmark.testdriver.AbstractParameterPool;
import benchmark.testdriver.LocalSPARQLParameterPool;
import benchmark.testdriver.Query;
import benchmark.testdriver.QueryMix;


/**
 * @author Olaf Hartig
 */
public class ModExecution extends ModGeneral
{
	final protected ArgDecl idirDecl = new ArgDecl( ArgDecl.HasValue, "idir" );
	final protected ArgDecl qdirDecl = new ArgDecl( ArgDecl.HasValue, "qdir" );
	final protected ArgDecl qseqDecl = new ArgDecl( ArgDecl.HasValue, "qseq" );
	final protected ArgDecl seedDecl = new ArgDecl( ArgDecl.HasValue, "seed" );

	static protected long dfltSeed = 5055L;

	private File inputParamDir;
	private AbstractParameterPool parameterPool;
	private File queryDir;
	private Integer[] queryNoSequence;
	private QueryMix queryMix;
	private long seed;


	// implementation of the ArgModuleGeneral interface

	public void registerWith ( CmdGeneral cmdLine )
	{
		super.registerWith( cmdLine );

		cmdLine.add( idirDecl,
		             "--idir",
		             "Directory with the input parameters as generated by BSBM (mandatory)") ;
		cmdLine.add( qdirDecl,
		             "--qdir",
		             "Directory with the query data (mandatory)") ;
		cmdLine.add( qseqDecl,
		             "--qseq",
		             "Number of the queries that make up the query mix (mandatory)") ;

		cmdLine.getUsage().startCategory( "Execution options" );

		cmdLine.add( seedDecl,
		             "--seed",
		             "A long number to seed the random query parameter generator (default: " + String.valueOf(dfltSeed) + ")") ;
	}


	// implementation of the ArgModule interface

	public void processArgs ( CmdArgModule cmdLine )
	{
		parameterPool = null;
		queryMix = null;

		super.processArgs ( cmdLine );

		if ( cmdLine.contains(idirDecl) )
		{
			String inputParamDirName = cmdLine.getValue( idirDecl );
			inputParamDir = new File( inputParamDirName );
			if ( ! inputParamDir.exists() || ! inputParamDir.isDirectory() ) {
				cmdLine.cmdError( "The given input param directory does not exist or not a directory" );
			}
		}
		else {
			cmdLine.cmdError( "No input param directory specified" );
		}

		if ( cmdLine.contains(qdirDecl) )
		{
			String queryDirName = cmdLine.getValue( qdirDecl );
			queryDir = new File( queryDirName );
			if ( ! queryDir.exists() || ! queryDir.isDirectory() ) {
				cmdLine.cmdError( "The given query directory does not exist or not a directory" );
			}
		}
		else {
			cmdLine.cmdError( "No query directory specified" );
		}

		if ( cmdLine.contains(qseqDecl) )
		{
			String queryNoSequenceString = cmdLine.getValue( qseqDecl );
			String[] queryNoSequenceStrings = queryNoSequenceString.split( " " );
			queryNoSequence = new Integer[ queryNoSequenceStrings.length ];
			for ( int i = 0; i < queryNoSequenceStrings.length; ++i )
			{
				try {
					queryNoSequence[i] = Integer.decode( queryNoSequenceStrings[i] );
				} catch ( NumberFormatException e ) {
					cmdLine.cmdError( "The given query sequence contains unsupported characters (" + e.getMessage() + ")" );
					break;
				}
			}
		}
		else {
			cmdLine.cmdError( "No query sequence specified" );
		}

		if ( cmdLine.contains(seedDecl) )
		{
			try {
				seed = Long.parseLong( cmdLine.getValue(seedDecl) );
			} catch ( NumberFormatException e ) {
				cmdLine.cmdError( "Reading the given seed failed: " + Utils.className(e) );
			}
		}
		else {
			seed = dfltSeed;
		}
	}


	// accessors

	public AbstractParameterPool getParameterPool ()
	{
		if ( parameterPool == null ) {
			parameterPool = new LocalSPARQLParameterPool( inputParamDir, seed );
		}

		return parameterPool;
	}

	public QueryMix getQueryMix ()
	{
		if ( queryMix == null )
		{
			int queryCount = 0;
			for ( int i = 0; i < queryNoSequence.length; ++i )
			{
				if ( queryNoSequence[i] > queryCount ) {
					queryCount = queryNoSequence[i];
				}
			}

			Query[] queries = new Query[queryCount];
			for ( int i = 0; i < queries.length; ++i ) {
				queries[i] = null;
			}

			for ( int i = 0; i < queryNoSequence.length; ++i )
			{
				int queryNo = queryNoSequence[i];
				if ( queries[queryNo-1] == null ) {
					queries[queryNo-1] = new Query( new File(queryDir, "query" + queryNo + ".txt"),
					                                new File(queryDir, "query" + queryNo + "desc.txt"),
					                                "%" );
				}
			}

			queryMix = new QueryMix( queries, queryNoSequence );
		}

		return queryMix;
	}

}