package bench;

import java.io.PrintStream;
import java.net.URL;
import java.util.Iterator;
import java.util.Set;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.util.IteratorCollection;

import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.ResultSet;

import de.fuberlin.wiwiss.ng4j.NamedGraph;
import de.fuberlin.wiwiss.ng4j.NamedGraphSet;

import org.squin.common.Statistics;
import org.squin.dataset.QueriedDataset;
import org.squin.dataset.jenacommon.QueriedDatasetWrappingJenaGraph;

import com.vladium.utils.ObjectProfiler;

import benchmark.testdriver.AbstractParameterPool;
import benchmark.testdriver.Query;
import benchmark.testdriver.QueryMix;


/**
 * Command line tool that executes the tRDF4Jena benchmark.
 *
 * @author Olaf Hartig
 */
public class execute extends CmdExecBase
{
	final protected ModBenchmark modBenchmark = new ModBenchmark();


	public static void main ( String [] args )
	{
		new execute(args).mainRun();
	}


	// initialization

	public execute ( String[] args )
	{
		super( args );
		addModule( modBenchmark );
	}


	// implementation of the CmdExecBase interface

	protected void execWorker( Model data, QueryMix queryMix, AbstractParameterPool parameterPool )
	{
		for ( int run = - modBenchmark.getW(); run < modBenchmark.getRuns(); ++run )
		{
			System.out.println( "Starting run " + String.valueOf(run) );
			queryMix.setRun( run );
			while( queryMix.hasNext() )
			{
				Query query = queryMix.getNext();
				query.setParameters( parameterPool.getParametersForQuery(query) );
				executeQuery( query, data );
			}
			queryMix.finishRun();
		}

		print( queryMix, System.out );
		int estSize = ObjectProfiler.sizeof(data);
		System.out.println( "Estimated size of the queried dataset: " + estSize );
		Statistics stats = null;
		if ( data.getGraph() instanceof QueriedDatasetWrappingJenaGraph ) {
			stats = ( (QueriedDatasetWrappingJenaGraph) data.getGraph() ).queriedDataset.getStatistics();
			System.out.println( "Statistics of the queried dataset: " + stats.toString() );
		}

		// measure the clean-up time by removing all RDF graphs sequentially from the queried dataset
		long cleanupTime = cleanup( data );
		System.out.println( "Clean-up time: " + cleanupTime + "ns" );

		System.out.print( queryMix.getCQET() + "," + queryMix.getQueryMixGeometricMean() + "," + estSize + "," + cleanupTime );
		if ( stats != null ) {
			System.out.print( "," + stats.toStringCSV() );
		}
		System.out.println();
	}


	// helpers

	protected void executeQuery ( Query q, Model data )
	{
		long startTime = System.nanoTime();
		int resultSize = 0;

		QueryExecution qe = QueryExecutionFactory.create( q.getQueryString(), data );
		if ( q.getQueryType().equals(Query.SELECT_TYPE) )
		{
			ResultSet results = qe.execSelect();
			while ( results.hasNext() )
			{
				++resultSize;
				results.nextSolution();
			}
		}
		else if ( q.getQueryType().equals(Query.CONSTRUCT_TYPE) )
		{
			Model results = qe.execConstruct();
		}
		else if ( q.getQueryType().equals(Query.DESCRIBE_TYPE) )
		{
			Model results = qe.execDescribe();
		}

		Long durationInMillies = System.nanoTime() - startTime;
		q.getQueryMix().setCurrent( resultSize, durationInMillies.doubleValue() / 1000000000 );
	}

	protected long cleanup ( Model data )
	{
		long cleanupTime;
		if ( data.getGraph() instanceof QueriedDatasetWrappingJenaGraph ) {
			QueriedDataset queriedDataset = ( (QueriedDatasetWrappingJenaGraph) data.getGraph() ).queriedDataset;
			Set<URL> graphURLs = queriedDataset.getRDFGraphsSourceURLs();

			long cleanupStartTime = System.nanoTime();
			for ( URL url : graphURLs ) {
				queriedDataset.removeRDFGraph( url );
			}
			cleanupTime = System.nanoTime() - cleanupStartTime;
		}
		else if ( data.getGraph() instanceof DatasetGraphWrappingJenaGraph ) {
			long cleanupStartTime = System.nanoTime();
			( (DatasetGraphWrappingJenaGraph) data.getGraph() ).dataset.close();
			cleanupTime = System.nanoTime() - cleanupStartTime;
		}
		else {
			NamedGraphSet ngs = null;
			if ( data.getGraph() instanceof IdBasedNamedGraphSetImpl.IdBasedUnionGraph ) {
				ngs = ( (IdBasedNamedGraphSetImpl.IdBasedUnionGraph) data.getGraph() ).ngs;
			}
			else if ( data.getGraph() instanceof NamedGraphSetImpl.UnionGraph ) {
				ngs = ( (NamedGraphSetImpl.UnionGraph) data.getGraph() ).ngs;
			}

			Set<NamedGraph> tmp = IteratorCollection.iteratorToSet( ngs.listGraphs() );
			long cleanupStartTime = System.nanoTime();
			for ( NamedGraph ng : tmp ) {
				ngs.removeGraph( ng.getGraphName() );
			}
			cleanupTime = System.nanoTime() - cleanupStartTime;
		}

		return cleanupTime;
	}

	protected void print ( QueryMix queryMix, PrintStream out )
	{
		Query[] queries = queryMix.getQueries();
		double[] min = queryMix.getQmin();
		double[] max = queryMix.getQmax();
		double[] am = queryMix.getAqet();
		double[] gm = queryMix.getGeoMean();
		int[] rpq = queryMix.getRunsPerQuery();
		int[] minResults = queryMix.getMinResults();
		int[] maxResults = queryMix.getMaxResults();

		for ( int i = 0; i < queries.length; ++i )
		{
			out.println( "Measures for query " + String.valueOf(i+1) );
			out.println( "------------------------" );
			out.println( String.valueOf(rpq[i]) + " times executed" );
			out.print( "Exec.times:" );
			out.printf( "\tarith.mean %.6f sec.", am[i] );
			out.printf( "\tgeom.mean %.6f sec.", gm[i] );
			out.printf( "\tmin %.6f sec.", min[i] );
			out.printf( "\tmax %.6f sec.", max[i] );
			out.println();
			out.print( "Result sizes:" );
			out.printf( "\tmin " + minResults[i] );
			out.printf( "\tmax " + maxResults[i] );
			out.println();
			out.println();
		}

		out.println( "Overall measures" );
		out.println( "------------------------" );
		out.print( "Query mix exec.times:" );
		out.printf( "\tarith.mean %.6f sec.", queryMix.getCQET() );
		out.printf( "\tgeom.mean %.6f sec.", queryMix.getQueryMixGeometricMean() );
		out.printf( "\tmin %.6f sec.", queryMix.getMinQueryMixRuntime() );
		out.printf( "\tmax %.6f sec.", queryMix.getMaxQueryMixRuntime() );
		out.println();
		out.printf( "Total exec.time: %.6f sec.", queryMix.getTotalRuntime() );
		out.println();

	}
}
